
/*

    Support: Threaded socket routines for reflector
    Copyright (C) 1999  A. Schiffler

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    To contact the author try these:
    Andreas Schiffler, 
    ZKM, Lorezstr. 19, 76135 Karlsruhe, Germany
    schiffler@zkm.de

*/

#include "refsocket.h"

/* Socket IO */
   
int sock_write (int sockfd, char *buf, size_t count)
{
 size_t bytes_sent = 0;
 int this_write;
       
 while (bytes_sent < count) {
  do {
   this_write = write(sockfd, buf, count - bytes_sent);
  } while ( (this_write < 0) && (errno == EINTR) );
  if (this_write <= 0) {
   return(this_write);
  } 
  bytes_sent += this_write;
  buf += this_write;
 }
 return(count);
}

int sock_read (int sockfd, char *buf, size_t count)
{
 size_t bytes_read = 0;
 int this_read;
 fd_set readers;
 struct timeval timeout;
 int pending;

 while (bytes_read < count) {
  do {
   /* Check status of connection */
   pending=0;
   while (pending==0) {
    FD_ZERO(&readers);
    FD_SET(sockfd,&readers);
    timeout.tv_sec=5;
    timeout.tv_usec=0;
    pending=select(FD_SETSIZE,&readers,0,0,&timeout);
    if (pending<0) {
     fprintf (stderr,"Connection closed ...\n");
     return(-1);
    }
   } 
   this_read = read(sockfd, buf, count - bytes_read);
  } while ( (this_read < 0) && (errno == EINTR) );
  if ((this_read==0) && (pending==1)) {
   fprintf (stderr,"Connection closed ...\n");
   return(-1);
  }    
  if (this_read <= 0) {
   return(this_read);
  } 
  bytes_read += this_read;
  buf += this_read;
 }
 return(count);
}
      
int sock_gets (int sockfd, char *str, size_t count)
{
 int bytes_read;
 int total_count = 0;
 char *current_position;
 char last_read = 0;
 fd_set readers;
 struct timeval timeout;
 int pending;
          
 current_position = str;
 while ((last_read != 10) && (last_read !=13)) {
   /* Check status of connection */
   pending=0;
   while (pending==0) {
    FD_ZERO(&readers);
    FD_SET(sockfd,&readers);
    timeout.tv_sec=5;
    timeout.tv_usec=0;
    pending=select(FD_SETSIZE,&readers,0,0,&timeout);
    if (pending<0) {
     fprintf (stderr,"Connection closed ...\n");
     return(-1);
    }
   } 
   bytes_read = read(sockfd, &last_read, 1);
   if ((bytes_read==0) && (pending==1)) {
    fprintf (stderr,"Connection closed ...\n");
    return(-1);
   }    
   if (bytes_read <= 0) {
    return bytes_read; 
   }
   if ( (total_count < count) && (last_read != 10) && (last_read !=13) ) {
    *current_position = last_read;
    current_position++;
    total_count++;
   }
 }
 if (count > 0) *current_position = 0;
 return total_count;
}

/* Network service thread                 */

int NetworkThread(struct netthreads *nt)
{
 int connected_socket;
 char inbuffer[BUFFER_LEN];
 char outbuffer[BUFFER_LEN];
 int connected;
 int my_textlines_head=0;
 int my_textlines_finish=0;
 int writeflag=0;
 int readlen;
  
 /* Transfer socket handle to local variable */
 shared_data_lock(nt);
 connected_socket=nt->connected_socket; 
 my_textlines_head=nt->textlines_head;
 nt->socketcount++;
 fprintf (stderr,"Currently %i connections.\n",nt->socketcount);
 shared_data_unlock(nt);
 /* Go into processing loop */
 connected=1;
 while (connected) {
  /* Read text */
  if ((readlen=sock_gets(connected_socket,inbuffer,BUFFER_LEN))<0) {
   connected=0;
  } else {
   if (readlen>0) {
    if (strncasecmp(inbuffer,"LOGOUT",6)==0) {
     fprintf (stderr,"Connection closed ...\n");
     connected=0;
    } else {
     if (strncasecmp(inbuffer,"POLL",4)==0) {
      writeflag=1;
      shared_data_lock(nt);
      my_textlines_finish=nt->textlines_head;
      shared_data_unlock(nt);
     } else {
      /* Process inbuffer - store textline in ringbuffer */
      shared_data_lock(nt);
      if (strncasecmp(inbuffer,"PUSH",4)==0) {
       /* push mode - no text returned */
       strncpy(nt->textlines[nt->textlines_head],&inbuffer[4],BUFFER_LEN);
       writeflag=0;
      } else {
       /* normal mode */
       strncpy(nt->textlines[nt->textlines_head],inbuffer,BUFFER_LEN);
       writeflag=1;
      }     
      my_textlines_finish=nt->textlines_head;
      nt->textlines_head = ((nt->textlines_head+1) % BUFFER_NUM);
      shared_data_unlock(nt);
     }
     
     if (writeflag) {
      /* Write outbuffers - all textlines from last to current head */
      while ((my_textlines_head != my_textlines_finish) && (connected)) {
       shared_data_lock(nt);
       strncpy(outbuffer,nt->textlines[my_textlines_head],(BUFFER_LEN-1));
       shared_data_unlock(nt);
       strcat(outbuffer,"\n");
       my_textlines_head = ((my_textlines_head+1) % BUFFER_NUM);
       /* Send data to network socket if we are not in push-mode */
       connected=(sock_write(connected_socket,outbuffer,strlen(outbuffer))>0);
      }
     
      /* Send end marker string */
      strcpy(outbuffer,"OVER\n");
      /* Send data to network socket */
      connected=(sock_write(connected_socket,outbuffer,strlen(outbuffer))>0);
     }
    } 
   }
  }   
 }
 /* Exit and goodby */
 close(connected_socket);
 shared_data_lock(nt);
 nt->socketcount--;
 fprintf (stderr,"Currently %i connections.\n",nt->socketcount);
 shared_data_unlock(nt);
 return(0);
}

int shared_data_lock(struct netthreads *nt)
{
 /* Lock shared variables */ 
 pthread_mutex_lock(&nt->network_mutex);
 
 return(0);
}

int shared_data_unlock(struct netthreads *nt)
{
 /* Unlock shared variables */ 
 pthread_mutex_unlock(&nt->network_mutex);
 
 return(0);
}

int start_network_socket(struct netthreads *nt, int connected_socket)
{
 pthread_t network_thread;

 /* Store socket handle */
 shared_data_lock(nt);
 nt->connected_socket=connected_socket;
 shared_data_unlock(nt);
 /* Start grabbing thread */
 if (pthread_create(&network_thread, NULL,(void *)NetworkThread,(void *)nt)==-1) {
  perror ("pthread_create()");
  return(-1);
 } else {
  return(0);
 } 
}

int init_network_socket(struct netthreads *nt)
{

 fprintf (stderr,"Initializing ...\n");

 nt->socketcount=0;
 nt->textlines_head=0;
 /* Initialize sycro mutex */
 if (pthread_mutex_init(&nt->network_mutex, NULL)==-1) {
  perror("pthread_mutex_init()");
  return(-1);
 }
 
 return(0);
}
